# binarypacker.rb

class BinaryPacker

  def BinaryPacker.little_endian?
    [1].pack('s') == "\001\000"
  end

  def BinaryPacker.big_endian?
    [1].pack('s') == "\000\001"
  end

  class << self
    alias newobj new
  end

  def BinaryPacker.new(fmt)
    parse(fmt)
  end

  def BinaryPacker.parse(fmt)
    newobj(compile(fmt))
  end

  def BinaryPacker.compile(fmt)
    fmt.split(/[;\n]/).reject {|s| s.strip.empty? }.map {|spec|
      if numspec = spec.slice!(/\[(?:\*|\d+)\]/)
        num = numspec.slice(/\*|\d+/)
      else
        num = ''
      end
      f = FORMAT[spec.gsub(/\s+/, '')] or
          raise ArgumentError, "unknown format: #{spec}"
      f + num
    }.join('')
  end

  FORMAT = {}
  declare_int = lambda {|base, signed, unsigned|
    unless signed == '-'
      FORMAT[base]            = signed
      FORMAT['s' + base]      = signed
      FORMAT['signed' + base] = signed
    end
    FORMAT['u' + base]        = unsigned
    FORMAT['unsigned' + base] = unsigned
  }
  # size:   system DEPENDENT
  # endian: system DEPENDENT
  declare_int[*%w(  char        c   C   )]
  declare_int[*%w(  short       s!  S!  )]
  declare_int[*%w(  int         i   I   )]
  declare_int[*%w(  long        l!  L!  )]
  declare_int[*%w(  longlong    q   Q   )]

  # size:   system independent
  # endian: system DEPENDENT
  declare_int[*%w(  byte        c   C   )]
  declare_int[*%w(  int8        c   C   )]
  declare_int[*%w(  int16       s   S   )]
  declare_int[*%w(  int32       l   L   )]
  declare_int[*%w(  int64       q   Q   )]

  # size:   system independent
  # endian: system independent
  declare_int[*%w(  int16little -   v   )]
  declare_int[*%w(  int32little -   V   )]
  declare_int[*%w(  int16big    -   n   )]
  declare_int[*%w(  int32big    -   N   )]

  # size:   system DEPENDENT
  # endian: system DEPENDENT
  FORMAT['float'] = 'f'
  FORMAT['double'] = 'd'

  # size:   system DEPENDENT
  # endian: system independent
  FORMAT['floatlittle'] = 'e'
  FORMAT['doublelittle'] = 'E'
  FORMAT['floatbig'] = 'g'
  FORMAT['doublebig'] = 'G'

  def initialize(fmt)
    @pack_format = fmt
    @item_length = nil
    @byte_length = nil
  end

  def pack(items)
    unless items.size >= item_length()
      raise ArgumentError, "too few items: #{items.size} for #{item_length()}"
    end
    items.pack(@pack_format)
  end

  def unpack(bin)
    unless bin.size >= byte_length()
      raise ArgumentError, "binary too short: #{bin.size} for #{byte_length()}"
    end
    bin.unpack(@pack_format)
  end

  def item_length
    @item_length ||= _item_length()
  end

  def _item_length
    len = 0
    @pack_format.scan(/(\w)!?(\*|\d+)/) do |fmt, n|
      len += num2int(n)
    end
    len
  end
  private :_item_length

  def byte_length
    @byte_length ||= _byte_length()
  end

  def _byte_length
    len = 0
    @pack_format.scan(/(\w)!?(\*|\d+)?/) do |fmt, n|
      len += [1].pack(fmt).length * num2int(n)
    end
    len
  end
  private :_byte_length

  private

  def num2int(n)
    case n
    when nil then 1
    when '*' then 0
    else          n.to_i
    end
  end

end
